using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

namespace PropertyChangedDemo
{
    public static class ExpressionHelper
    {
        //Go through an expression and pick out all the unique MemberAccesses
        //e.g. if an expression is like l.Quantity * l.UnitPrice then it would return
        //["Quantity", "UnitPrice"] (actually the MemberInfo, not strings, which includes type information).
        public static IEnumerable<MemberInfo> GetSourceProperties(Expression exp)
        {
            return GetAllSourceProperties(exp).Distinct();
        }

        //Go through an expression and pick out all the MemberAccesses
        //e.g. if an expression is like l.Quantity * l.UnitPrice then it would return
        //["Quantity", "UnitPrice"] (actually the MemberInfo, not strings, which includes type information)
        private static IEnumerable<MemberInfo> GetAllSourceProperties(Expression exp)
        {
            if (exp.NodeType == ExpressionType.MemberAccess)
            {
                yield return ((MemberExpression)exp).Member;
            }
            else if (exp is ConditionalExpression)
            {
                var c = exp as ConditionalExpression;
                foreach (var o in GetAllSourceProperties(c.Test))
                    yield return o;
                foreach (var o in GetAllSourceProperties(c.IfFalse))
                    yield return o;
                foreach (var o in GetAllSourceProperties(c.IfTrue))
                    yield return o;
            }
            else if (exp is BinaryExpression)
            {
                foreach (var o in GetAllSourceProperties(((BinaryExpression)exp).Left))
                    yield return o;
                foreach (var o in GetAllSourceProperties(((BinaryExpression)exp).Right))
                    yield return o;
            }
            else if (exp is UnaryExpression)
            {
                foreach (var o in GetAllSourceProperties(((UnaryExpression)exp).Operand))
                    yield return o;
            }
            else if (exp.NodeType == ExpressionType.Lambda)
            {
                foreach (var o in GetAllSourceProperties(((LambdaExpression)exp).Body))
                    yield return o;
            }
            else if (exp.NodeType == ExpressionType.Call)
            {
                foreach (var argExpression in ((MethodCallExpression)exp).Arguments)
                {
                    foreach (var o in GetAllSourceProperties(argExpression))
                        yield return o;
                }
            }
        }

        public static MemberInfo GetMemberInfo(Expression exp)
        {
            switch (exp.NodeType)
            {
                case ExpressionType.Lambda:
                    return GetMemberInfo(((LambdaExpression)exp).Body);
                case ExpressionType.Convert:
                    return ((MemberExpression)((UnaryExpression)exp).Operand).Member;
                case ExpressionType.MemberAccess:
                    return ((MemberExpression)exp).Member;
                default:
                    throw new ArgumentException();
            }
       }
    }
}