﻿namespace PropertyChangedDemo
{
    //This is a model for a basic form that has two calculated properties:
    //1.  Subtotal = UnitPrice * Quantity
    //2.  Total = Subtotal + Freight
    public class Model : BaseModel
    {
        public Model()
        {
            //Initialize some dummy data so the form makes more sense at startup.
            _unitPrice = 9.99M;
            _quantity = 2M;
            _subtotal = 19.98M;
            _freight = 5M;
            _total = 24.98M;
        }

        //These are the basic properties for UnitPrice, Quantity, and Freight.  There's some boilerplate code for RaiseOnPropertyChanged.
        private decimal _unitPrice; public decimal UnitPrice { get { return _unitPrice; } set { if (_unitPrice != value) { _unitPrice = value; RaiseOnPropertyChanged("UnitPrice"); } } }  
        private decimal _quantity; public decimal Quantity { get { return _quantity; } set { if (_quantity != value) { _quantity = value; RaiseOnPropertyChanged("Quantity"); } } }
        private decimal _freight; public decimal Freight { get { return _freight; } set { if (_freight != value) { _freight = value; RaiseOnPropertyChanged("Freight"); } } }

        //This is a calculated property for the Subtotal.
        //The first line is the boilerplate property as before.
        //The second line is what does the declarative definition of the calculated property.
        private decimal _subtotal; public decimal Subtotal { get { return _subtotal; } set { if (_subtotal != value) { _subtotal = value; RaiseOnPropertyChanged("Subtotal"); } } }
        public static readonly BasicPropertyInfo SubtotalProperty = For<Model>(x => x.Subtotal).Define(x => x.UnitPrice * x.Quantity);

        private decimal _total; public decimal Total { get { return _total; } set { if (_total != value) { _total = value; RaiseOnPropertyChanged("Total"); } } }
        public static readonly BasicPropertyInfo TotalProperty = For<Model>(x => x.Total).Define(x => x.Subtotal + x.Freight);

        //In order for the declarative definition to work, we need to hook up to the PropertyChanged notifications
        //This could be done here in this model, in the form, or in some other helper class, depending on what makes sense in your application.
        protected override void RaiseOnPropertyChanged(string propertyName)
        {
            base.RaiseOnPropertyChanged(propertyName);
            //PropertyChangedActionManager tracks the calculated properties and actions that need to be done for each property change.
            //We're using the one for the Model class.
            //DoActions takes the instance of Model that's changing (so that it can access the source properties) and the name of the property being changed.
            PropertyChangedActionManager<Model>.Instance.DoActions(this, propertyName);
        }
    }
}
