using System;
using System.Collections.Generic;
using System.Linq;

namespace PropertyChangedDemo
{
    //NOTE:
    //There are two copies of this class:
    //- one with one generic parameter for a class listening to PropertyChanged events from itself
    //- another with two generic parameters for a class listening to PropertyChanged events from itself and another source object
    //Most likely any changes you make to one class will need to be made to the other one as well.

    #region PropertyChangedActionManager for when listening to the same object

    /// <summary>
    /// This class is used to make it easier to update actions based on PropertyChangedNotifications
    /// </summary>
    /// <typeparam name="TChanging"></typeparam>
    /// <typeparam name="TListener"></typeparam>
    public class PropertyChangedActionManager<TListener>
    {
        private Dictionary<string, List<Action<TListener>>> actions = new Dictionary<string, List<Action<TListener>>>();

        private static readonly PropertyChangedActionManager<TListener> _Instance = new PropertyChangedActionManager<TListener>();

        private PropertyChangedActionManager() { }
        public static PropertyChangedActionManager<TListener> Instance
        {
            get { return _Instance; }
        }

        public void RegisterListenerChangedAction(Action<TListener> action, params string[] updateFrom)
        {
            foreach (var p in updateFrom)
            {
                if (!actions.ContainsKey(p))
                    actions[p] = new List<Action<TListener>>();
                actions[p].Add(action);
            }
        }

        public void DoActions(TListener listening, string changedPropertyName)
        {
            if (listening != null)
            {
                var matchingProperty = actions.Keys.FirstOrDefault<string>(x => x == changedPropertyName);
                if (matchingProperty != null)
                {
                    foreach (var a in actions[matchingProperty])
                    {
                        a(listening);
                    }
                }
            }
        }
    }

    #endregion

    #region Class with two generic type parameters

    /// <summary>
    /// This class is used to make it easier to update actions based on PropertyChangedNotifications
    /// </summary>
    /// <typeparam name="TSource"></typeparam>
    /// <typeparam name="TListener"></typeparam>
    public class PropertyChangedActionManager<TSource, TListener>
    {
        private Dictionary<string, List<Action<TSource, TListener>>> sourceChangedActions = new Dictionary<string, List<Action<TSource, TListener>>>();
        private Dictionary<string, List<Action<TSource, TListener>>> listenerChangedActions = new Dictionary<string, List<Action<TSource, TListener>>>();

        private static readonly PropertyChangedActionManager<TSource, TListener> _Instance = new PropertyChangedActionManager<TSource, TListener>();

        private PropertyChangedActionManager() { }
        public static PropertyChangedActionManager<TSource, TListener> Instance
        {
            get { return _Instance; }
        }

        public void RegisterSourceChangedAction(Action<TSource, TListener> action, params string[] updateFrom)
        {
            foreach (var p in updateFrom)
            {
                if (!sourceChangedActions.ContainsKey(p))
                    sourceChangedActions[p] = new List<Action<TSource, TListener>>();
                sourceChangedActions[p].Add(action);
            }
        }

        public void RegisterListenerChangedAction(Action<TSource, TListener> action, params string[] updateFrom)
        {
            foreach (var p in updateFrom)
            {
                if (!listenerChangedActions.ContainsKey(p))
                    listenerChangedActions[p] = new List<Action<TSource, TListener>>();
                listenerChangedActions[p].Add(action);
            }
        }

        public void DoActions(TSource source, TListener listener)
        {
            foreach (var actionList in listenerChangedActions.Values.Union(sourceChangedActions.Values))
                foreach (var action in actionList)
                    action(source, listener);
        }

        public void DoActions(TSource source, TListener listener, string changedPropertyName)
        {
            if (source != null && listener != null)
            {
                DoListenerChangedActions(source, listener, changedPropertyName);
                DoSourceChangedActions(source, listener, changedPropertyName);
            }
        }

        private void DoListenerChangedActions(TSource source, TListener listener, string changedPropertyName)
        {
            DoChangedActions(source, listener, changedPropertyName, listenerChangedActions);
        }

        private void DoSourceChangedActions(TSource source, TListener listener, string changedPropertyName)
        {
            DoChangedActions(source, listener, changedPropertyName, sourceChangedActions);
        }

        private static void DoChangedActions(TSource source, TListener listener, string changedPropertyName, Dictionary<string, List<Action<TSource, TListener>>> dict)
        {
            var matchingProperty = dict.Keys.FirstOrDefault<string>(x => x == changedPropertyName);
            if (matchingProperty != null)
            {
                foreach (var a in dict[matchingProperty])
                    a(source, listener);
            }
        }
    }

    #endregion

}